// ============================================================================
// COPYRIGHT NOTICE
// ----------------------------------------------------------------------------
// (This is the open source ISC license, see
// http://en.wikipedia.org/wiki/ISC_license
// for more info)
//
// Copyright © 2020-2023  Andreas M. Rammelt <rammi@caff.de>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//=============================================================================
// Latest version on https://caff.de/projects/decaff-commons/
//=============================================================================
package de.caff.generics;

import de.caff.annotation.NotNull;
import de.caff.generics.function.BooleanFunction2;
import de.caff.generics.function.BooleanOperator2;
import de.caff.generics.function.BooleanPredicate2;
import de.caff.generics.tuple.Tuple;
import de.caff.generics.tuple.Tuple2;

import java.io.Serializable;
import java.util.Locale;
import java.util.Objects;

/**
 * A pair of primitive boolean values.
 * <p>
 * This is similar to {@link Pair}, but for raw {@code boolean} values.
 * As these values are defined final, this class is also immutable.
 * @author <a href="mailto:rammi@caff.de">Rammi</a>
 * @since December 04, 2020
 */
public final class BooleanPair
        implements Serializable,
                   BooleanIndexable
{
  private static final long serialVersionUID = -4169359896148258597L;
  /** The first value of this pair. */
  public final boolean first;
  /** The second value of this pair. */
  public final boolean second;

  /**
   * Constructor.
   * @param first    first value of pair
   * @param second   second value of pair
   */
  public BooleanPair(boolean first, boolean second)
  {
    this.first = first;
    this.second = second;
  }

  @Override
  public int size()
  {
    return 2;
  }

  @Override
  public boolean get(int index)
  {
    switch (index) {
    case 0:
      return first;
    case 1:
      return second;
    default:
      throw new IndexOutOfBoundsException(String.format("Pair has only 2 elements, so %d is out of bounds!", index));
    }
  }

  /**
   * Apply a 2-boolean function on this pair.
   * @param function function to apply, {@link #first} becomes its first, {@link #second} its second argument
   * @param <T> output type
   * @return result of applying the function
   */
  public <T> T applyAsBooleans(@NotNull BooleanFunction2<T> function)
  {
    return function.applyAsBoolean(first, second);
  }

  /**
   * Call an 2-boolean operator on this pair.
   * @param operator operator to apply, {@link #first} becomes its first, {@link #second} its second argument
   * @return result of applying the operator
   */
  public boolean operate(@NotNull BooleanOperator2 operator)
  {
    return operator.applyAsBoolean(first, second);
  }

  /**
   * Call a 2-boolean predicate on this pair. 
   * @param predicate predicate to apply, {@link #first} becomes its first, {@link #second} its second argument 
   * @return result of testing this pair
   */
  public boolean test(@NotNull BooleanPredicate2 predicate)
  {
    return predicate.testBools(first, second);
  }

  /**
   * Convert this into a tuple.
   * @return tuple created from this pair
   */
  @NotNull
  public Tuple2<Boolean, Boolean> toTuple()
  {
    return Tuple.of(first, second);
  }

  @Override
  public boolean equals(Object o)
  {
    if (this == o) {
      return true;
    }
    if (o == null || getClass() != o.getClass()) {
      return false;
    }
    final BooleanPair that = (BooleanPair)o;
    return that.first == first && that.second == second;
  }

  @Override
  public int hashCode()
  {
    return Objects.hash(first, second);
  }

  @Override
  @NotNull
  public String toString()
  {
    return String.format(Locale.US, "BooleanPair(%s, %s)", first, second);
  }
}
