// ============================================================================
// COPYRIGHT NOTICE
// ----------------------------------------------------------------------------
// (This is the open source ISC license, see
// http://en.wikipedia.org/wiki/ISC_license
// for more info)
//
// Copyright © 2020-2023  Andreas M. Rammelt <rammi@caff.de>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//=============================================================================
// Latest version on https://caff.de/projects/decaff-commons/
//=============================================================================
package de.caff.generics;

import de.caff.annotation.NotNull;
import de.caff.generics.function.ByteFunction2;
import de.caff.generics.function.ByteOperator2;
import de.caff.generics.function.BytePredicate2;
import de.caff.generics.tuple.Tuple;
import de.caff.generics.tuple.Tuple2;

import java.io.Serializable;
import java.util.Locale;
import java.util.Objects;

import static de.caff.generics.Primitives.unsigned;

/**
 * A pair of primitive byte values.
 * <p>
 * This is similar to {@link Pair}, but for raw {@code byte} values.
 * As these values are defined final, this class is also immutable.
 * @author <a href="mailto:rammi@caff.de">Rammi</a>
 * @since December 04, 2020
 */
public final class BytePair
        implements Serializable,
                   ByteIndexable
{
  private static final long serialVersionUID = 4196797097671990594L;
  /** The first value of this pair. */
  public final byte first;
  /** The second value of this pair. */
  public final byte second;

  /**
   * Constructor.
   * @param first    first value of pair
   * @param second   second value of pair
   */
  public BytePair(byte first, byte second)
  {
    this.first = first;
    this.second = second;
  }

  /**
   * Convert this pair of {@code byte}s into a pair of {@code int}s.
   * @return int pair with same values
   */
  @NotNull
  public IntPair toIntPair()
  {
    return new IntPair(first, second);
  }

  /**
   * Convert this pair of {@code byte}s into a pair of {@code int}s assuming unsigned values.
   * @return int pair with the values of this pair assumed unsigned
   */
  @NotNull
  public IntPair toUnsignedIntPair()
  {
    return new IntPair(unsigned(first),
                       unsigned(second));
  }

  /**
   * Get this pair with a different first value.
   * @param first new first value
   * @return pair with the given {@code first} value,
   *         and the {@link #second} value of this pair
   */
  @NotNull
  public BytePair withFirst(byte first)
  {
    if (first == this.first) {
      return this;
    }
    return new BytePair(first, this.second);
  }

  /**
   * Get this pair with a different first value.
   * @param second new second value
   * @return pair with the {@link #first} value of this pair,
   *         and the given {@code second} value
   */
  @NotNull
  public BytePair withSecond(byte second)
  {
    if (second == this.second) {
      return this;
    }
    return new BytePair(this.first, second);
  }

  /**
   * Get this pair with swapped values.
   * @return pair with swapped {@link #first} and {@link #second} values
   */
  @NotNull
  public BytePair swapped()
  {
    if (first == second) {
      return this;
    }
    return new BytePair(second, first);
  }

  @Override
  public int size()
  {
    return 2;
  }

  @Override
  public byte get(int index)
  {
    switch (index) {
    case 0:
      return first;
    case 1:
      return second;
    default:
      throw new IndexOutOfBoundsException(String.format("Pair has only 2 elements, so %d is out of bounds!", index));
    }
  }

  /**
   * Apply a 2-byte function on this pair.
   * @param function function to apply, {@link #first} becomes its first, {@link #second} its second argument
   * @param <T> output type
   * @return result of applying the function
   */
  public <T> T applyAsBytes(@NotNull ByteFunction2<T> function)
  {
    return function.applyAsByte(first, second);
  }

  /**
   * Call an 2-byte operator on this pair.
   * @param operator operator to apply, {@link #first} becomes its first, {@link #second} its second argument
   * @return result of applying the operator
   */
  public byte operate(@NotNull ByteOperator2 operator)
  {
    return operator.applyAsByte(first, second);
  }

  /**
   * Call a 2-byte predicate on this pair. 
   * @param predicate predicate to apply, {@link #first} becomes its first, {@link #second} its second argument 
   * @return result of testing this pair
   */
  public boolean test(@NotNull BytePredicate2 predicate)
  {
    return predicate.testBytes(first, second);
  }

  /**
   * Convert this into a tuple.
   * @return tuple created from this pair
   */
  @NotNull
  public Tuple2<Byte, Byte> toTuple()
  {
    return Tuple.of(first, second);
  }

  @Override
  public boolean equals(Object o)
  {
    if (this == o) {
      return true;
    }
    if (o == null || getClass() != o.getClass()) {
      return false;
    }
    final BytePair that = (BytePair)o;
    return that.first == first && that.second == second;
  }

  @Override
  public int hashCode()
  {
    return Objects.hash(first, second);
  }

  @Override
  @NotNull
  public String toString()
  {
    return String.format(Locale.US, "BytePair(%s, %s)", first, second);
  }
}
