// ============================================================================
// COPYRIGHT NOTICE
// ----------------------------------------------------------------------------
// (This is the open source ISC license, see
// http://en.wikipedia.org/wiki/ISC_license
// for more info)
//
// Copyright © 2020-2023  Andreas M. Rammelt <rammi@caff.de>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//=============================================================================
// Latest version on https://caff.de/projects/decaff-commons/
//=============================================================================
package de.caff.generics;

import de.caff.annotation.NotNull;
import de.caff.generics.function.ByteConsumer;

import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.PrimitiveIterator;
import java.util.function.Consumer;

/**
 * An iterator specialized for {@code short} values.
 * <p>
 * This is missing in the standard library: an iterator
 * for raw {@code byte} values which does everything what
 * {@link OfInt} does for {@code int}s.
 *
 * @author <a href="mailto:rammi@caff.de">Rammi</a>
 * @since September 27, 2020
 */
public interface PrimitiveByteIterator
        extends PrimitiveIterator<Byte, ByteConsumer>
{
  /** Empty iterator of primitive byte values. */
  PrimitiveByteIterator EMPTY = new PrimitiveByteIterator()
  {
    @Override
    public byte nextByte()
    {
      throw new NoSuchElementException("Empty byte iterator has no elements!");
    }

    @Override
    public boolean hasNext()
    {
      return false;
    }

    @Override
    public void forEachRemaining(@NotNull ByteConsumer action)
    {
    }

    @Override
    public void forEachRemaining(@NotNull Consumer<? super Byte> action)
    {
    }
  };

  /**
   * Returns the next {@code byte} element in the iteration.
   *
   * @return the next {@code byte} element in the iteration
   * @throws NoSuchElementException if the iteration has no more elements
   */
  byte nextByte();

  /**
   * Performs the given action for each remaining element until all elements
   * have been processed or the action throws an exception.  Actions are
   * performed in the order of iteration, if that order is specified.
   * Exceptions thrown by the action are relayed to the caller.
   *
   * <p>The default implementation behaves as if:
   * <pre>{@code
   *     while (hasNext())
   *         action.accept(nextByte());
   * }</pre>
   *
   * @param action The action to be performed for each element
   * @throws NullPointerException if the specified action is null
   */
  default void forEachRemaining(@NotNull ByteConsumer action) {
    Objects.requireNonNull(action);
    while (hasNext()) {
      action.accept(nextByte());
    }
  }

  /**
   * {@inheritDoc}
   * The default implementation boxes the result of calling
   * {@link #nextByte()}, and returns that boxed result.
   */
  @Override
  default Byte next() {
    return nextByte();
  }

  /**
   * {@inheritDoc}
   * If the action is an instance of {@code ByteConsumer} then it is cast
   * to {@code ByteConsumer} and passed to {@link #forEachRemaining};
   * otherwise the action is adapted to an instance of
   * {@code ByteConsumer}, by boxing the argument of {@code ByteConsumer},
   * and then passed to {@link #forEachRemaining}.
   */
  @Override
  default void forEachRemaining(@NotNull Consumer<? super Byte> action) {
    if (action instanceof ByteConsumer) {
      forEachRemaining((ByteConsumer) action);
    }
    else {
      // The method reference action::accept is never null
      Objects.requireNonNull(action);
      forEachRemaining((ByteConsumer) action::accept);
    }
  }

  /**
   * View this byte iterator as a short iterator.
   * Please note that you should use either this iterator
   * of the returned one, but don't mix using both.
   * @return short iterator view of this iterator
   */
  @NotNull
  default PrimitiveShortIterator asShortIterator()
  {
    return new PrimitiveShortIterator()
    {
      @Override
      public short nextShort()
      {
        return PrimitiveByteIterator.this.nextByte();
      }

      @Override
      public boolean hasNext()
      {
        return PrimitiveByteIterator.this.hasNext();
      }
    };
  }

  /**
   * View this byte iterator as an int iterator.
   * Please note that you should use either this iterator
   * of the returned one, but don't mix using both.
   * @return int iterator view of this iterator
   */
  @NotNull
  default PrimitiveIterator.OfInt asIntIterator()
  {
    return new OfInt()
    {
      @Override
      public int nextInt()
      {
        return PrimitiveByteIterator.this.nextByte();
      }

      @Override
      public boolean hasNext()
      {
        return PrimitiveByteIterator.this.hasNext();
      }
    };
  }

  /**
   * View this byte iterator as an int iterator using the values of
   * this iterator as unsigned bytes.
   * Please note that you should use either this iterator
   * of the returned one, but don't mix using both.
   * @return int iterator view of this iterator
   */
  @NotNull
  default PrimitiveIterator.OfInt asUnsignedIntIterator()
  {
    return new OfInt()
    {
      @Override
      public int nextInt()
      {
        return PrimitiveByteIterator.this.nextByte() & 0xff;
      }

      @Override
      public boolean hasNext()
      {
        return PrimitiveByteIterator.this.hasNext();
      }
    };
  }

  /**
   * View this byte iterator as a long iterator.
   * Please note that you should use either this iterator
   * of the returned one, but don't mix using both.
   * @return long iterator view of this iterator
   */
  @NotNull
  default PrimitiveIterator.OfLong asLongIterator()
  {
    return new OfLong()
    {
      @Override
      public long nextLong()
      {
        return PrimitiveByteIterator.this.nextByte();
      }

      @Override
      public boolean hasNext()
      {
        return PrimitiveByteIterator.this.hasNext();
      }
    };
  }

  /**
   * View this byte iterator as a long iterator using the values of
   * this iterator as unsigned bytes.
   * Please note that you should use either this iterator
   * of the returned one, but don't mix using both.
   * @return long iterator view of this iterator
   */
  @NotNull
  default PrimitiveIterator.OfLong asUnsignedLongIterator()
  {
    return new OfLong()
    {
      @Override
      public long nextLong()
      {
        return PrimitiveByteIterator.this.nextByte() & 0xffL;
      }

      @Override
      public boolean hasNext()
      {
        return PrimitiveByteIterator.this.hasNext();
      }
    };
  }
}
