// ============================================================================
// COPYRIGHT NOTICE
// ----------------------------------------------------------------------------
// (This is the open source ISC license, see
// http://en.wikipedia.org/wiki/ISC_license
// for more info)
//
// Copyright © 2016-2023  Andreas M. Rammelt <rammi@caff.de>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//=============================================================================
// Latest version on https://caff.de/projects/decaff-commons/
//=============================================================================
package de.caff.generics;

import de.caff.annotation.NotNull;
import de.caff.generics.function.FragileShortConsumer;
import de.caff.generics.function.ShortConsumer;

import java.util.Iterator;

/**
 * Iterable which iterates over a primitive short.
 * @author <a href="mailto:rammi@caff.de">Rammi</a>
 * @since December 13, 2021
 * @see ShortIndexable for creating one on the fly
 */
public interface PrimitiveShortIterable
        extends Iterable<Short>
{
  /**
   * Empty iterator of primitive short values.
   * @see PrimitiveShortIterator#EMPTY
   */
  PrimitiveShortIterator EMPTY_SHORT_ITERATOR = PrimitiveShortIterator.EMPTY;
  /** Empty iterable of primitive short values. */
  PrimitiveShortIterable EMPTY = () -> EMPTY_SHORT_ITERATOR;

  /**
   * Get a primitive short iterator.
   * @return short iterator
   */
  @NotNull
  PrimitiveShortIterator shortIterator();

  @Override
  @NotNull
  default Iterator<Short> iterator()
  {
    return shortIterator();
  }

  /**
   * Invoke the given action with each short value in this iterable.
   * @param action action to be call for each short value in sequemce.
   */
  default void forEachShort(@NotNull ShortConsumer action)
  {
    for (PrimitiveShortIterator it = shortIterator();  it.hasNext(); ) {
      action.accept(it.nextShort());
    }
  }

  /**
   * Invoke the given fragile action with each value in this iterable.
   * @param action action to be called for each value in sequence
   * @param <E> checked exception thrown by action
   * @throws E if {@code action} throws it
   */
  default <E extends Exception> void forEachShortFragile(@NotNull FragileShortConsumer<E> action)
          throws E
  {
    for (PrimitiveShortIterator it = shortIterator(); it.hasNext(); ) {
      action.accept(it.nextShort());
    }
  }

  /**
   * Does this short iterable contain the given value?
   * This is an operation which takes time linear to the size of this iterable
   * and should only be used rarely.
   * @param value value to check for
   * @return {@code true} if the value is contained in this iterable, {@code false} if not
   */
  default boolean containsShort(short value)
  {
    for (PrimitiveShortIterator iterator = shortIterator();
         iterator.hasNext();  ) {
      if (iterator.nextShort() == value) {
        return true;
      }
    }
    return false;
  }
  /**
   * View this byte iterable as an int iterable.
   * @return int iterable view of this byte iterable
   */
  @NotNull
  default PrimitiveIntIterable asIntIterable()
  {
    return () -> shortIterator().asIntIterator();
  }

  /**
   * View this byte iterable as an int iterable which uses the unsigned byte values
   * of this iterable.
   * @return unsigned int iterable view of this byte iterable
   */
  @NotNull
  default PrimitiveIntIterable asUnsignedIntIterable()
  {
    return () -> shortIterator().asUnsignedIntIterator();
  }

  /**
   * View this byte iterable as a long iterable.
   * @return long iterable view of this byte iterable
   */
  @NotNull
  default PrimitiveLongIterable asLongIterable()
  {
    return () -> shortIterator().asLongIterator();
  }

  /**
   * View this byte iterable as a long iterable which uses the unsigned byte values
   * of this iterable.
   * @return unsigned long iterable view of this byte iterable
   */
  @NotNull
  default PrimitiveLongIterable asUnsignedLongIterable()
  {
    return () -> shortIterator().asUnsignedLongIterator();
  }
}
