// ============================================================================
// COPYRIGHT NOTICE
// ----------------------------------------------------------------------------
// (This is the open source ISC license, see
// http://en.wikipedia.org/wiki/ISC_license
// for more info)
//
// Copyright © 2012-2023  Andreas M. Rammelt <rammi@caff.de>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//=============================================================================
// Latest version on https://caff.de/projects/decaff-commons/
//=============================================================================
package de.caff.generics;

import de.caff.annotation.NotNull;

/**
 * A uniform matcher is a {@link Matcher} which can only compare
 * objects of the same type.
 * <p>
 * Besides {@link #IDENTICAL} which only accepts the same object as identical,
 * useful standard implementations are also
 * {@code Objects::equals} or {@code Objects::deepEquals}.
 *
 * @author <a href="mailto:rammi@caff.de">Rammi</a>
 * @see de.caff.generics.HashCodeCalculator
 */
@FunctionalInterface
public interface UniformMatcher<T>
        extends Matcher<T, T>
{
  /** Uniform matcher which only accepts identical objects. */
  UniformMatcher<Object> IDENTICAL = (v1, v2) -> v1 == v2;

  /**
   * Identical matcher.
   * This just returns {@link #IDENTICAL} but may be used in cases when
   * the interface does not accept matchers with a super type.
   * @param <V> value type
   * @return uniform matcher which only accepts identical objects
   */
  @NotNull
  @SuppressWarnings("unchecked")  // because everything is handled at object level
  default
  <V>
  UniformMatcher<V> identical()
  {
    return (UniformMatcher<V>)IDENTICAL;
  }
}
